import pystokes.interface
import pystokes.periodic
import pystokes.twoWalls
import pystokes.unbounded
import pystokes.utils
import pystokes.wallBounded
import pystokes.forceFields
import pystokes.phoretic.unbounded
import pystokes.phoretic.wallBounded



## latest version of PyStokes
__version__="2.2.5"
